# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from math import isnan

from mo_dots import datas
from mo_dots import lists
from mo_dots.datas import *
from mo_dots.fields import *
from mo_dots.lists import *
from mo_dots.nones import *
from mo_dots.objects import DataObject, DataClass, object_to_data
from mo_dots.utils import *
from mo_dots.utils import _null_types as null_types

__all__ = [
    "coalesce",
    "concat_field",
    "Data",
    "DataClass",
    "DataObject",
    "dict_to_data",
    "endswith_field",
    "exists",
    "FlatList",
    "from_data",
    "get_attr",
    "hash_value",
    "inverse",
    "is_container",
    "is_data",
    "is_finite",
    "is_list",
    "is_many",
    "is_missing",
    "is_not_null",
    "is_null",
    "is_primitive",
    "is_sequence",
    "join_field",
    "last",
    "leaves",
    "leaves_to_data",
    "list_to_data",
    "listwrap",
    "literal_field",
    "missing",
    "Null",
    "NullType",
    "null_types",
    "object_to_data",
    "PATH_NOT_FOUND",
    "relative_field",
    "register_data",
    "register_many",
    "register_list",
    "register_primitive",
    "register_type",
    "set_attr",
    "set_default",
    "split_field",
    "startswith_field",
    "tail_field",
    "to_data",
    "tuplewrap",
    "unliteral_field",
    "unwrap",
    "unwraplist",
]


_module_type = type(sys.modules[__name__])
_builtin_zip = zip
_get = object.__getattribute__
_set = object.__setattr__
_new = object.__new__
_dict_zip = zip


def inverse(d):
    """
    reverse the k:v pairs
    """
    output = {}
    for k, v in from_data(d).items():
        output[v] = output.get(v, [])
        output[v].append(k)
    return output


def coalesce(*args):
    # pick the first not null value
    # http://en.wikipedia.org/wiki/Null_coalescing_operator
    for a in args:
        if a != None:
            return to_data(a)
    return Null


def zip(keys, values):
    """
    CONVERT LIST OF KEY/VALUE PAIRS TO Data
    PLEASE `import mo_dots`, AND CALL `mo_dots.zip()`
    """
    output = Data()
    for k, v in _dict_zip(keys, values):
        output[k] = v
    return output


def missing(value):
    raise NotImplementedError("use is_missing")


def fromkeys(keys, value=None):
    if is_null(value):
        return Data()
    return dict_to_data(dict.fromkeys(keys, value))


def set_default(d, *dicts):
    """
    RECURSIVE MERGE OF MULTIPLE dicts MOST IMPORTANT FIRST

    UPDATES d WITH THE MERGE RESULT, WHERE MERGE RESULT IS DEFINED AS:
    FOR EACH LEAF, RETURN THE FIRST NOT-NULL LEAF VALUE

    :param dicts: dicts IN PRIORITY ORDER, HIGHEST TO LOWEST
    :return: d
    """
    agg = d if d or is_data(d) else {}
    for p in dicts:
        _set_default(agg, p, seen={})
    return to_data(agg)


def _set_default(d, default, seen=None):
    """
    ANY VALUE NOT SET WILL BE SET BY THE default
    THIS IS RECURSIVE
    """
    if default is None:
        return

    for k, default_value in default.items():
        raw_value = from_data(default_value)  # TWO DIFFERENT Dicts CAN SHARE id() BECAUSE THEY ARE SHORT LIVED
        if is_data(d):
            existing_value = d.get(k)
        else:
            existing_value = _get_attr(d, [k])

        if is_null(existing_value):
            if default_value != None:
                if is_data(default_value):
                    df = seen.get(id(raw_value))
                    if df is not None:
                        _set_attr(d, [k], df)
                    else:
                        copy_dict = {}
                        seen[id(raw_value)] = copy_dict
                        _set_attr(d, [k], copy_dict)
                        _set_default(copy_dict, default_value, seen)
                else:
                    # ASSUME PRIMITIVE (OR LIST, WHICH WE DO NOT COPY)
                    try:
                        _set_attr(d, [k], default_value)
                    except Exception as e:
                        if PATH_NOT_FOUND not in e:
                            get_logger().error("Can not set attribute {{name}}", name=k, cause=e)
        elif is_list(existing_value) or is_list(default_value):
            _set_attr(d, [k], None)
            _set_attr(d, [k], listwrap(existing_value) + listwrap(default_value))
        elif (hasattr(existing_value, "__setattr__") or is_data(existing_value)) and is_data(default_value):
            df = seen.get(id(raw_value))
            if df is not None:
                _set_attr(d, [k], df)
            else:
                seen[id(raw_value)] = existing_value
                _set_default(existing_value, default_value, seen)


def _getdefault(obj, key):
    """
    obj ANY OBJECT
    key IS EXPECTED TO BE LITERAL (NO ESCAPING)
    TRY BOTH ATTRIBUTE AND ITEM ACCESS, OR RETURN Null
    """
    try:
        return obj[key]
    except Exception as f:
        pass

    if is_sequence(obj):
        return [_getdefault(o, key) for o in obj]

    try:
        if _get(obj, CLASS) is not dict:
            return getattr(obj, key)
    except Exception as f:
        pass

    try:
        if float(key) == round(float(key), 0):
            return obj[int(key)]
    except Exception as f:
        pass

    # TODO: FIGURE OUT WHY THIS WAS EVER HERE (AND MAKE A TEST)
    # try:
    #     return eval("obj."+str(key))
    # except Exception as f:
    #     pass
    return NullType(obj, key)


PATH_NOT_FOUND = "Path not found"
AMBIGUOUS_PATH_FOUND = "Path is ambiguous"


def set_attr(obj, path, value):
    """
    SAME AS object.__setattr__(), BUT USES DOT-DELIMITED path
    RETURN OLD VALUE
    """
    try:
        return _set_attr(obj, split_field(path), value)
    except Exception as cause:
        Log = get_logger()
        if PATH_NOT_FOUND in cause:
            Log.warning(PATH_NOT_FOUND + ": {{path}}", path=path, cause=cause)
        else:
            Log.error("Problem setting value", cause=cause)


def get_attr(obj, path):
    """
    SAME AS object.__getattr__(), BUT USES DOT-DELIMITED path
    """
    try:
        return _get_attr(obj, split_field(path))
    except Exception as cause:
        Log = get_logger()
        if PATH_NOT_FOUND in cause:
            Log.error(PATH_NOT_FOUND + ": {{path}}", path=path, cause=cause)
        else:
            Log.error("Problem setting value", cause=cause)


def _get_attr(obj, path):
    if not path:
        return obj

    attr_name = path[0]

    if isinstance(obj, _module_type):
        if attr_name in obj.__dict__:
            return _get_attr(obj.__dict__[attr_name], path[1:])
        elif attr_name in dir(obj):
            return _get_attr(obj[attr_name], path[1:])

        # TRY FILESYSTEM
        File = get_module("mo_files").File
        possible_error = None
        python_file = (File(obj.__file__).parent / attr_name).set_extension("py")
        python_module = File(obj.__file__).parent / attr_name / "__init__.py"
        if python_file.exists or python_module.exists:
            try:
                # THIS CASE IS WHEN THE __init__.py DOES NOT IMPORT THE SUBDIR FILE
                # WE CAN STILL PUT THE PATH TO THE FILE IN THE from CLAUSE
                if len(path) == 1:
                    # GET MODULE OBJECT
                    output = __import__(
                        obj.__name__ + str(".") + str(attr_name), globals(), locals(), [str(attr_name)], 0,
                    )
                    return output
                else:
                    # GET VARIABLE IN MODULE
                    output = __import__(
                        obj.__name__ + str(".") + str(attr_name), globals(), locals(), [str(path[1])], 0,
                    )
                    return _get_attr(output, path[1:])
            except Exception as e:
                Except = get_module("mo_logs.exceptions.Except")
                possible_error = Except.wrap(e)

        # TRY A CASE-INSENSITIVE MATCH
        matched_attr_name = lower_match(attr_name, dir(obj))
        if not matched_attr_name:
            get_logger().warning(
                PATH_NOT_FOUND + "({{name|quote}}) Returning None.", name=attr_name, cause=possible_error,
            )
        elif len(matched_attr_name) > 1:
            get_logger().error(AMBIGUOUS_PATH_FOUND + " {{paths}}", paths=attr_name)
        else:
            return _get_attr(obj, matched_attr_name + path[1:])

    try:
        obj = obj[int(attr_name)]
        return _get_attr(obj, path[1:])
    except Exception:
        pass

    try:
        obj = getattr(obj, attr_name)
        return _get_attr(obj, path[1:])
    except Exception:
        pass

    try:
        obj = obj[attr_name]
        return _get_attr(obj, path[1:])
    except Exception as f:
        return NullType(obj, attr_name)


def _set_attr(obj_, path, value):
    obj = _get_attr(obj_, path[:-1])
    if obj is None:
        # DELIBERATE USE OF `is`: WE DO NOT WHAT TO CATCH Null HERE (THEY CAN BE SET)
        get_logger().error(PATH_NOT_FOUND + " tried to get attribute of None")

    attr_name = path[-1]

    # ACTUAL SETTING OF VALUE
    try:
        old_value = _get_attr(obj, [attr_name])
        old_type = _get(old_value, CLASS)
        if is_null(old_value) or is_primitive(old_value):
            old_value = None
            new_value = value
        elif is_null(value):
            new_value = None
        else:
            new_value = _get(old_value, CLASS)(value)  # TRY TO MAKE INSTANCE OF SAME CLASS
    except Exception:
        old_value = None
        new_value = value

    try:
        setattr(obj, attr_name, new_value)
        return old_value
    except Exception as e:
        try:
            obj[attr_name] = new_value
            return old_value
        except Exception as f:
            get_logger().error(PATH_NOT_FOUND, cause=[f, e])


def lower_match(value, candidates):
    return [v for v in candidates if v.lower() == value.lower()]


def to_data(v=None) -> object:
    """
    WRAP AS Data OBJECT FOR DATA PROCESSING: https://github.com/klahnakoski/mo-dots/tree/dev/docs
    :param v:  THE VALUE TO WRAP
    :return:  Data INSTANCE
    """

    type_ = _get(v, CLASS)

    if type_ in (dict, OrderedDict):
        m = _new(Data)
        _set(m, SLOT, v)
        return m
    elif type_ is none_type:
        return Null
    elif type_ is tuple:
        return list_to_data(v)
    elif type_ is list:
        return list_to_data(v)
    elif type_ in generator_types:
        return list_to_data(list(from_data(vv) for vv in v))
    else:
        return v


wrap = to_data


def from_data(v):
    if v is None:
        return None
    _type = _get(v, CLASS)
    if _type is NullType:
        return None
    elif _type is Data:
        d = _get(v, SLOT)
        return d
    elif _type is FlatList:
        return _get(v, SLOT)
    elif _type is DataObject:
        return _get(v, SLOT)
    elif _type in generator_types:
        return (from_data(vv) for vv in v)
    elif _type is float:
        if isnan(v):
            return None
        return v
    return v


unwrap = from_data


def listwrap(value):
    """
    PERFORMS THE FOLLOWING TRANSLATION
    None -> []
    value -> [value]
    [...] -> [...]  (unchanged list)

    ## MOTIVATION ##
    OFTEN IT IS NICE TO ALLOW FUNCTION PARAMETERS TO BE ASSIGNED A VALUE,
    OR A list-OF-VALUES, OR NULL.  CHECKING FOR WHICH THE CALLER USED IS
    TEDIOUS.  INSTEAD WE CAST FROM THOSE THREE CASES TO THE SINGLE CASE
    OF A LIST

    # BEFORE
    def do_it(a):
        if a is None:
            return
        if not isinstance(a, list):
            a=[a]
        for x in a:
            # do something

    # AFTER
    def do_it(a):
        for x in listwrap(a):
            # do something

    """
    if is_null(value):
        return FlatList()
    elif is_list(value):
        if isinstance(value, list):
            return list_to_data(value)
        else:
            return value
    elif is_many(value):
        return list_to_data(list(value))
    else:
        return list_to_data([from_data(value)])


def unwraplist(v):
    """
    LISTS WITH ZERO AND ONE element MAP TO None AND element RESPECTIVELY
    """
    if is_list(v):
        if len(v) == 0:
            return None
        elif len(v) == 1:
            return from_data(v[0])
        else:
            return from_data(v)
    else:
        return from_data(v)


def tuplewrap(value):
    """
    INTENDED TO TURN lists INTO tuples FOR USE AS KEYS
    """
    if is_null(value):
        return tuple()
    elif is_many(value):
        return tuple(tuplewrap(v) if is_sequence(v) else v for v in value)
    else:
        return (from_data(value),)


# LEGACY PROPERTIES
class _DeferManyTypes:

    @cache
    def warning(self):
        get_logger().warning("DEPRECATED: Use mo_dots.utils._many_types", stack_depth=2)

    def __iter__(self):
        yield from utils._many_types
setattr(lists, '_many_types', _DeferManyTypes())
setattr(lists, 'many_types', _DeferManyTypes())
setattr(utils, 'many_types', _DeferManyTypes())



class _DeferDataTypes:

    @cache
    def warning(self):
        get_logger().warning("DEPRECATED: Use mo_dots.utils._data_types", stack_depth=2)

    def __iter__(self):
        self.warning()
        yield from utils._data_types
setattr(datas, '_data_types', _DeferDataTypes())
setattr(datas, 'data_types', _DeferDataTypes())


# EXPORT
export("mo_dots.datas", to_data)
export("mo_dots.datas", from_data)
export("mo_dots.datas", coalesce)
export("mo_dots.datas", _getdefault)
export("mo_dots.datas", listwrap)

export("mo_dots.lists", to_data)
export("mo_dots.lists", from_data)
export("mo_dots.lists", coalesce)
export("mo_dots.lists", get_attr)

export("mo_dots.nones", to_data)
export("mo_dots.nones", get_attr)

export("mo_dots.objects", to_data)
export("mo_dots.objects", from_data)
export("mo_dots.objects", get_attr)
export("mo_dots.objects", set_attr)
export("mo_dots.objects", set_default)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
import importlib
import types
from collections import OrderedDict
from dataclasses import is_dataclass
from datetime import datetime, date, timedelta, time
from decimal import Decimal

from mo_future import none_type, generator_types

from mo_dots.logging import get_logger

KEY = "_key"
SLOT = "_internal_value"
CLASS = "__class__"
_get = object.__getattribute__


def get_module(name):
    try:
        return importlib.import_module(name)
    except Exception as e:
        get_logger().error(
            "`pip install " + name.split(".")[0].replace("_", "-") + "` to enable this feature", cause=e,
        )


_null_types = (none_type,)


def register_null_type(_type):
    global _null_types
    _null_types = tuple(set(_null_types + (_type,)))


def is_null(value):
    # RETURN True IF EFFECTIVELY NOTHING
    _class = _get(value, CLASS)
    if _class in _null_types:
        return True
    return False


def is_not_null(value):
    _class = _get(value, CLASS)
    if _class in _null_types:
        return False
    return True


def is_missing(t) -> bool:
    # RETURN True IF EFFECTIVELY NOTHING
    return not t and isinstance(t, (str, *_null_types, *_many_types))


def exists(value) -> bool:
    return not is_missing(value)


_primitive_types = (
    str,
    bytes,
    int,
    float,
    bool,
    Decimal,
    datetime,
    date,
    time,
    timedelta,
    dict.values.__class__,
    object.__class__,
    types.FunctionType,
    types.MethodType,
)


def is_primitive(value):
    return isinstance(value, _primitive_types)


def register_primitive(_type):
    global _primitive_types
    _primitive_types = tuple(set(_primitive_types + (_type,)))


_data_types = data_types = (dict, OrderedDict)  # TYPES TO HOLD DATA


def register_data(type_):
    """
    :param type_:  ADD OTHER TYPE THAT HOLDS DATA
    :return:
    """
    global _data_types
    _data_types = tuple(set(_data_types + (type_,)))


def is_data(d):
    """
    :param d:
    :return: True IF d IS A TYPE THAT HOLDS DATA
    """
    return _get(d, CLASS) in _data_types


_known_data_types = tuple()


def register_type(*_classes):
    global _known_data_types
    _known_data_types = tuple(set(_known_data_types + _classes))


def is_namedtuple(obj):
    return isinstance(obj, tuple) and hasattr(obj, "_fields")


def is_data_object(obj):
    return isinstance(obj, _known_data_types) or is_namedtuple(obj) or is_dataclass(obj)


def is_known_data_type(_class):
    return _class in _known_data_types


list_types = (list,)
container_types = (list, set)
finite_types = (list, set, tuple)
sequence_types = (list, tuple) + generator_types
_many_types = tuple(set(list_types + container_types + sequence_types))


def register_list(_type):
    # lists belong to all categories
    global list_types, container_types, finite_types, sequence_types, _many_types
    list_types = tuple(set(list_types + (_type,)))
    container_types = tuple(set(container_types + (_type,)))
    finite_types = tuple(set(finite_types + (_type,)))
    sequence_types = tuple(set(sequence_types + (_type,)))
    _many_types = tuple(set(_many_types + (_type,)))


# ITERATORS THAT ARE CONSIDERED PRIMITIVE
not_many_names = ("str", "unicode", "binary", "NullType", "NoneType", "dict", "Data")


def is_list(l):
    # ORDERED, AND CAN CHANGE CONTENTS
    return _get(l, CLASS) in list_types


def is_container(l):
    # CAN ADD AND REMOVE ELEMENTS
    return _get(l, CLASS) in container_types


def is_sequence(l):
    # HAS AN ORDER, INCLUDES GENERATORS
    return _get(l, CLASS) in sequence_types


def is_finite(l):
    # CAN PERFORM len(l); NOT A GENERATOR
    return _get(l, CLASS) in finite_types


def is_many(value):
    # REPRESENTS MULTIPLE VALUES
    # TODO: CLEAN UP THIS LOGIC
    # THIS IS COMPLICATED BECAUSE I AM UNSURE ABOUT ALL THE "PRIMITIVE TYPES"
    # I WOULD LIKE TO POSITIVELY CATCH many_types, BUT MAYBE IT IS EASIER TO DETECT: Iterable, BUT NOT PRIMITIVE
    # UNTIL WE HAVE A COMPLETE SLOT, WE KEEP ALL THIS warning() CODE
    global _many_types
    if isinstance(value, _many_types):
        return True

    type_ = _get(value, CLASS)
    if issubclass(type_, types.GeneratorType):
        _many_types = _many_types + (type_,)
        get_logger.warning("is_many() can not detect generator {type}", type=type_.__name__)
        return True
    return False


def register_many(_type):
    global _many_types
    _many_types = _many_types + (_type,)


def cache(func):
    """
    DECORATOR TO CACHE THE RESULT OF A FUNCTION
    """
    cache = {}

    def wrapper(*args):
        if args in cache:
            return cache[args]
        else:
            result = func(*args)
            cache[args] = result
            return result

    return wrapper



import re
import sys

from mo_future import generator_types, flatten
from mo_imports import expect

from mo_dots.utils import get_logger, CLASS, is_missing, is_null

_module_type = type(sys.modules[__name__])
_builtin_zip = zip
_get = object.__getattribute__


ESCAPE_DOTS1 = re.compile(r"(^\.|\.$)")  # DOTS AT START/END
ESCAPE_DOTS2 = re.compile(r"(?<!^)\.(?!$)")  # INTERNAL DOTS
ILLEGAL_DOTS = re.compile(r"[^.]\.(?:\.\.)+")  # ODD DOTS ARE NOT ALLOWED
SPLIT_DOTS = re.compile(r"(?<!\.)\.(?!\.)")  # SINGLE DOTS
UNESCAPE_DOTS = re.compile(r"\x08|(?:\.\.)")  # ENCODED DOTS


def literal_field(field):
    """
    RETURN SAME WITH DOTS (`.`) ESCAPED
    """
    try:
        return ESCAPE_DOTS2.sub("..", ESCAPE_DOTS1.sub("\b", field))
    except Exception as e:
        get_logger().error("bad literal", e)


def unliteral_field(field):
    """
    DUE TO PATHOLOGY IN MY CODE WE HAVE A path WITH ESCAPED DOTS BUT WE WANT OT USE IT ON A dict, NOT A Data
    a = dict()
    b = Data(a)
    a[unliteral_field(k)]==b[k] (for all k)

    :param field: THE STRING TO DE-literal IZE
    :return: SIMPLER STRING
    """
    return UNESCAPE_DOTS.sub(".", field)


def tail_field(field):
    """
    RETURN THE FIRST STEP IN PATH, ALONG WITH THE REMAINING TAILf
    IN (first, rest) PAIR
    """
    if field == "." or is_missing(field):
        return ".", "."
    elif "." in field:
        path = split_field(field)
        if path[0].startswith("."):
            return path[0], join_field(path[1:])
        return literal_field(path[0]), join_field(path[1:])
    else:
        return field, "."


def split_field(field):
    """
    RETURN field AS ARRAY OF DOT-SEPARATED FIELDS
    """
    if ILLEGAL_DOTS.search(field):
        get_logger().error("Odd number of dots is not allowed")
    if field.startswith(".."):
        remainder = field.lstrip(".")
        back = len(field) - len(remainder) - 1
        return [".."] * back + [UNESCAPE_DOTS.sub(".", k) for k in SPLIT_DOTS.split(remainder) if k]
    else:
        return [UNESCAPE_DOTS.sub(".", k) for k in SPLIT_DOTS.split(field) if k]


def join_field(path):
    """
    RETURN field SEQUENCE AS STRING
    """
    if _get(path, CLASS) in generator_types:
        path = list(path)

    if not path:
        return "."

    prefix = ""
    while True:
        try:
            i = path.index("..")
            if i == 0:
                prefix += "."
                path = path[1:]
            else:
                path = path[: i - 1] + path[i + 1 :]
        except ValueError:
            return ("." if prefix else "") + prefix + ".".join(literal_field(f) for f in path)


def concat_field(*fields):
    return join_field(flatten(split_field(f) for f in fields))


def startswith_field(field, prefix):
    """
    RETURN True IF field PATH STRING STARTS WITH prefix PATH STRING
    """
    if prefix == None:
        return False
    if prefix.startswith("."):
        return True
        # f_back = len(field) - len(field.strip("."))
        # p_back = len(prefix) - len(prefix.strip("."))
        # if f_back > p_back:
        #     return False
        # else:
        #     return True

    if field.startswith(prefix):
        lp = len(prefix)
        if len(field) == len(prefix) or field[lp] in (".", "\b") and field[lp + 1] not in (".", "\b"):
            return True
    return False


def endswith_field(field, suffix):
    """
    RETURN True IF field PATH STRING ENDS WITH suffix PATH STRING
    """
    if is_null(suffix):
        return False
    if suffix == ".":
        return True

    if field.endswith(suffix):
        ls = len(suffix)
        if len(field) == ls or field[-ls - 1] in (".", "\b") and field[-ls - 2] not in (".", "\b"):
            return True
    return False


def relative_field(field, parent):
    """
    RETURN field PATH WITH RESPECT TO parent
    """
    if parent == ".":
        return field

    field_path = split_field(field)
    parent_path = split_field(parent)
    common = 0
    for f, p in _builtin_zip(field_path, parent_path):
        if f != p:
            break
        common += 1

    tail = join_field(field_path[common:])
    if len(parent_path) <= common:
        return join_field(field_path[common:])

    dots = "." * (len(parent_path) - common)
    if tail == ".":
        return "." + dots
    else:
        return "." + dots + tail



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from copy import copy, deepcopy
from decimal import Decimal

from mo_future import generator_types, MutableMapping, first
from mo_imports import expect, export

from mo_dots import utils
from mo_dots.fields import split_field, literal_field, concat_field
from mo_dots.nones import Null, NullType
from mo_dots.utils import *

(
    _getdefault,
    coalesce,
    listwrap,
    from_data,
    to_data,
    list_to_data,
    FlatList,
    DataObject,
    get_keys,
    object_to_data,
) = expect(
    "_getdefault",
    "coalesce",
    "listwrap",
    "from_data",
    "to_data",
    "list_to_data",
    "FlatList",
    "DataObject",
    "get_keys",
    "object_to_data",
)


_get = object.__getattribute__
_set = object.__setattr__
_new = object.__new__

DEBUG = False


class Data:
    """
    Please see https://github.com/klahnakoski/mo-dots/tree/dev/docs#data-replaces-pythons-dict
    """

    __slots__ = [SLOT]

    def __init__(self, *args, **kwargs):
        """
        CONSTRUCT DATA WITH GIVEN PROPERTY VALUES
        """
        if args:
            raise Exception("only keywords are allowed, not " + args[0].__class__.__name__)
        _set(self, SLOT, kwargs)

    def __bool__(self):
        d = _get(self, SLOT)
        if _get(d, CLASS) is dict:
            return True
        else:
            return d != None

    __nonzero__ = __bool__

    def __contains__(self, item):
        value = Data.__getitem__(self, item)
        if is_data(value) or value:
            return True
        return False

    def __iter__(self):
        d = _get(self, SLOT)
        if _get(d, CLASS) is dict:
            yield from d.items()
        else:
            yield from d.__iter__()

    def __getitem__(self, key):
        if is_null(key):
            return Null
        if key == ".":
            output = _get(self, SLOT)
            if is_data(output):
                return self
            else:
                return output

        key = str(key)
        d = _get(self, SLOT)

        if key.find(".") >= 0:
            seq = _split_field(key)
            for n in seq:
                if _get(d, CLASS) is NullType:
                    d = NullType(d, n)  # OH DEAR, Null TREATS n AS PATH, NOT LITERAL
                elif is_many(d):
                    d = [_getdefault(dd, n) for dd in d]
                else:
                    d = _getdefault(d, n)  # EVERYTHING ELSE TREATS n AS LITERAL

            return to_data(d)
        else:
            o = d.get(key)

        if is_null(o):
            return NullType(d, key)
        return to_data(o)

    def __setitem__(self, key, value):
        if key == "":
            get_logger().error("key is empty string.  Probably a bad idea")
        if is_null(key):
            return Null
        if key == ".":
            # SOMETHING TERRIBLE HAPPENS WHEN value IS NOT A Mapping;
            # HOPEFULLY THE ONLY OTHER METHOD RUN ON self IS from_data()
            v = from_data(value)
            if is_many(v):
                _set(self, CLASS, FlatList)
            _set(self, SLOT, v)
            return self
        try:
            d = _get(self, SLOT)
            value = from_data(value)
            if "." not in key:
                if value is None:
                    d.pop(key, None)
                else:
                    d[key] = value
                return self

            seq = _split_field(key)
            for k in seq[:-1]:
                d = _getdefault(d, k)
            if is_null(value):
                try:
                    d.pop(seq[-1], None)
                except Exception as _:
                    pass
            elif is_null(d):
                d[literal_field(seq[-1])] = value
            elif is_sequence(d):
                for dd in d:
                    from_data(dd)[seq[-1]] = value
            else:
                d[seq[-1]] = value
            return self
        except Exception as e:
            from mo_logs import Log

            Log.error("can not set key={{key}}", key=key, cause=e)

    def __getattr__(self, key):
        d = _get(self, SLOT)
        v = d.get(key)
        t = _get(v, CLASS)

        # OPTIMIZED to_data()
        if t is dict:
            return dict_to_data(v)
        elif t in utils._null_types:
            return NullType(d, key)
        elif t is list:
            return list_to_data(v)
        elif t in generator_types:
            return FlatList(list(from_data(vv) for vv in v))
        else:
            return v

    def __setattr__(self, key, value):
        d = _get(self, SLOT)
        value = from_data(value)
        if value is None:
            d = _get(self, SLOT)
            d.pop(key, None)
        else:
            d[key] = value
        return self

    def __add__(self, other):
        return _iadd(_iadd({}, self), other)

    def __radd__(self, other):
        return _iadd(_iadd({}, other), self)

    def __iadd__(self, other):
        return _iadd(self, other)

    def __or__(self, other):
        """
        RECURSIVE COALESCE OF DATA PROPERTIES
        """
        if not is_data(other):
            get_logger().error("Expecting Data")

        d = _get(self, SLOT)
        output = Data(**d)  # COPY
        output.__ior__(other)
        return output

    def __ror__(self, other):
        """
        RECURSIVE COALESCE OF DATA PROPERTIES
        """
        if not is_data(other):
            get_logger().error("Expecting Data")

        return to_data(other).__or__(self)

    def __ior__(self, other):
        """
        RECURSIVE COALESCE OF DATA PROPERTIES
        """
        d = _get(self, SLOT)
        if not is_data(other):
            if is_missing(d) or (isinstance(d, dict) and not d):
                _set(self, SLOT, other)
            return self

        for ok, ov in other.items():
            if is_null(ov):
                continue

            sv = to_data(d.get(ok))
            if is_null(sv):
                d[ok] = ov
            elif is_data(sv):
                d[ok] = sv | ov
        return self

    def __hash__(self):
        d = _get(self, SLOT)
        return hash_value(d)

    def __eq__(self, other):
        if self is other:
            return True

        d = _get(self, SLOT)
        if _get(d, CLASS) is not dict:
            return d == other

        if not d and is_null(other):
            return False

        if not is_data(other):
            return False
        e = other
        for k, v in d.items():
            if e.get(k) != v:
                return False
        for k, v in e.items():
            if d.get(k) != v:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def get(self, key, default=Null):
        v = self[key]
        if _get(v, CLASS) == NullType:
            if default is Null:
                return NullType(self, key)
            return default
        return v

    def items(self):
        d = _get(self, SLOT)
        return [(k, to_data(v)) for k, v in d.items() if v != None or is_data(v)]

    def leaves(self, prefix=None):
        """
        LIKE items() BUT RECURSIVE, AND ONLY FOR THE LEAVES (non dict) VALUES
        """
        return leaves(self, prefix)

    def iteritems(self):
        # LOW LEVEL ITERATION, NO WRAPPING
        d = _get(self, SLOT)
        return ((k, to_data(v)) for k, v in d.items())

    def pop(self, key, default=Null):
        if is_null(key):
            return Null
        if key == ".":
            raise NotImplemented()

        key = str(key)
        d = _get(self, SLOT)

        if key.find(".") >= 0:
            seq = _split_field(key)
            for n in seq[:-1]:
                if _get(d, CLASS) is NullType:
                    d = NullType(d, n)  # OH DEAR, Null TREATS n AS PATH, NOT LITERAL
                elif is_many(d):
                    d = [_getdefault(dd, n) for dd in d]
                else:
                    d = _getdefault(d, n)  # EVERYTHING ELSE TREATS n AS LITERAL
            key = seq[-1]

        o = d.get(key)
        if is_null(o):
            if default is Null:
                return NullType(d, key)
            return default

        d[key] = None
        return to_data(o)

    def keys(self):
        d = _get(self, SLOT)
        return set(d.keys())

    def values(self):
        d = _get(self, SLOT)
        return listwrap(list(d.values()))

    def clear(self):
        get_logger().error("clear() not supported")

    def __len__(self):
        d = _get(self, SLOT)
        return dict.__len__(d)

    def copy(self):
        d = _get(self, SLOT)
        if _get(d, CLASS) is dict:
            return Data(**d)
        else:
            return copy(d)

    def __copy__(self):
        d = _get(self, SLOT)
        if _get(d, CLASS) is dict:
            return Data(**self)
        else:
            return copy(d)

    def __deepcopy__(self, memo):
        d = _get(self, SLOT)
        return to_data(deepcopy(d, memo))

    def __delitem__(self, key):
        if "." not in key:
            d = _get(self, SLOT)
            d.pop(key, None)
            return

        d = _get(self, SLOT)
        seq = _split_field(key)
        for k in seq[:-1]:
            d = d[k]
        d.pop(seq[-1], None)

    def __delattr__(self, key):
        key = str(key)
        d = _get(self, SLOT)
        d.pop(key, None)

    def setdefault(self, k, d=None):
        v = self[k]
        if is_null(v):
            self[k] = d
            return d
        return v

    def __str__(self):
        return str(_get(self, SLOT))

    def __dir__(self):
        d = _get(self, SLOT)
        return d.keys()

    def __repr__(self):
        try:
            return f"to_data({repr(_get(self, SLOT))})"
        except Exception as e:
            return "Data(?)"


MutableMapping.register(Data)
register_data(Data)

def leaves(value, prefix=None):
    """
    LIKE items() BUT RECURSIVE, AND ONLY FOR THE LEAVES (non dict) VALUES
    SEE leaves_to_data FOR THE INVERSE

    :param value: THE Mapping TO TRAVERSE
    :param prefix:  OPTIONAL PREFIX GIVEN TO EACH KEY
    :return: Data, WHICH EACH KEY BEING A PATH INTO value TREE
    """
    if not prefix:
        yield from _leaves(".", value, tuple())
    else:
        for k, v in _leaves(".", value, tuple()):
            yield prefix + k, v


def _leaves(parent, value, path):
    val = from_data(value)
    _id = id(val)
    if _id in path:
        yield parent, value
        return
    obj = object_to_data(val)
    if obj is val or is_many(val):
        yield parent, value
        return

    for k in get_keys(obj):
        try:
            v = obj[literal_field(k)]
            if is_missing(v):
                continue
            kk = concat_field(parent, literal_field(k))
            vv = object_to_data(v)
            yield from _leaves(kk, vv, path + (_id,))
        except Exception as cause:
            get_logger().error("Do not know how to handle", cause=cause)


def _split_field(field):
    """
    SIMPLE SPLIT, NO CHECKS
    """
    return [k.replace("\b", ".") for k in field.replace("..", "\b").split(".")]


def _iadd(self, other):
    """
    RECURSIVE ADDITION OF DATA PROPERTIES
    * LISTS ARE CONCATENATED
    * SETS ARE UNIONED
    * NUMBERS ARE ADDED
    """

    if not is_data(other):
        # HAPPENS WHEN _iadd WITH ['.'] SELF REFERENCE
        d = _get(self, SLOT)
        if isinstance(d, dict) and not len(d):
            # LOOKS LIKE A FRESH Data OBJECT (AN IDENTITY ELEMENT)
            # ∀ x, x += {} => x
            d = Data()
        else:
            d = dict_to_data({"$": self})
        d += dict_to_data({"$": other})
        d["."] = d["$"]
        return d

    d = from_data(self)
    for ok, ov in other.items():
        sv = d.get(ok)
        if is_null(sv):
            d[ok] = from_data(deepcopy(ov))
        elif isinstance(ov, (Decimal, float, int)):
            if is_data(sv):
                get_logger().error(
                    "can not add {{stype}} with {{otype}",
                    stype=_get(sv, CLASS).__name__,
                    otype=_get(ov, CLASS).__name__,
                )
            elif is_many(sv):
                d[ok].append(ov)
            else:
                d[ok] = sv + ov
        elif is_many(ov):
            d[ok] = from_data(listwrap(sv) + ov)
        elif is_data(ov):
            if is_data(sv):
                _iadd(sv, ov)
            elif is_many(sv):
                d[ok].append(ov)
            else:
                get_logger().error(
                    "can not add {{stype}} with {{otype}",
                    stype=_get(sv, CLASS).__name__,
                    otype=_get(ov, CLASS).__name__,
                )
        else:
            if is_data(sv):
                get_logger().error(
                    "can not add {{stype}} with {{otype}",
                    stype=_get(sv, CLASS).__name__,
                    otype=_get(ov, CLASS).__name__,
                )
            else:
                d[ok].append(ov)
    return self


def hash_value(v):
    if is_many(v):
        return hash_value(first(v))
    elif is_data(v):
        return hash_value(first(v.values()))
    else:
        return hash(v)


def dict_to_data(d):
    """
    FASTEST WAY TO MAKE Data, DO NOT CHECK TYPE
    :param d: dict
    :return: Data
    """
    m = _new(Data)
    _set(m, SLOT, d)
    return m


def leaves_to_data(value):
    """
    dict WITH DOTS IN KEYS IS INTERPRETED AS A PATH
    """
    return to_data(_leaves_to_data(value))


wrap_leaves = leaves_to_data


def _leaves_to_data(value):
    """
    RETURN UNWRAPPED STRUCTURES
    """
    if is_null(value):
        return None

    if is_primitive(value):
        return value

    _class = _get(value, CLASS)
    if _class in utils._data_types:
        if _class is Data:
            value = from_data(value)

        output = {}
        for key, value in value.items():
            value = _leaves_to_data(value)

            if key == "":
                get_logger().error("key is empty string.  Probably a bad idea")

            seq = split_field(key)
            if not seq:
                if not output:
                    output = value
                continue
            if not is_data(output):
                output = {}
            d = output
            for k in seq[:-1]:
                e = d.get(k, None)
                if not is_data(e):
                    e = d[k] = {}
                d = e

            if is_null(value):
                d.pop(seq[-1], None)
            else:
                d[seq[-1]] = value
        return output

    if hasattr(value, "__iter__"):
        return [_leaves_to_data(v) for v in value]

    return value



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from collections import OrderedDict
from copy import deepcopy

from mo_future import generator_types, get_function_arguments, get_function_defaults, Mapping
from mo_imports import export, expect

from mo_dots.datas import Data, _iadd, dict_to_data
from mo_dots.lists import FlatList, list_to_data
from mo_dots.nones import NullType, Null
from mo_dots.utils import CLASS, SLOT, register_data, get_logger, is_primitive, is_known_data_type, is_null, \
    register_type

get_attr, set_attr, to_data, from_data, set_default = expect(
    "get_attr", "set_attr", "to_data", "from_data", "set_default"
)

_new = object.__new__
_get = object.__getattribute__
_set = object.__setattr__


_known_fields = {}  #  map from type to field names
ignored_attributes = set(dir(object)) | set(dir(dict.values.__class__))


class DataObject(Mapping):
    """
    TREAT AN OBJECT LIKE DATA
    """

    __slots__ = [SLOT]

    def __init__(self, obj):
        _set(self, SLOT, obj)

    def __getattr__(self, item):
        obj = _get(self, SLOT)
        output = get_attr(obj, item)
        return object_to_data(output)

    def __setattr__(self, key, value):
        obj = _get(self, SLOT)
        set_attr(obj, key, value)

    def __getitem__(self, item):
        obj = _get(self, SLOT)
        output = get_attr(obj, item)
        return object_to_data(output)

    def __or__(self, other):
        return set_default({}, self, other)

    def __ror__(self, other):
        return to_data(other) | self

    def __add__(self, other):
        return to_data(_iadd(_iadd({}, self), other))

    def __radd__(self, other):
        return to_data(_iadd(_iadd({}, other), self))

    def get(self, item):
        obj = _get(self, SLOT)
        output = get_attr(obj, item)
        return object_to_data(output)

    def keys(self):
        return get_keys(self)

    def items(self):
        keys = self.keys()
        try:
            for k in keys:
                yield k, self[k]
        except Exception as cause:
            get_logger().error("problem with items", cause=cause)

    def __deepcopy__(self, memodict={}):
        output = {}
        for k, v in self.items():
            output[k] = from_data(deepcopy(v))
        return dict_to_data(output)

    def __data__(self):
        return self

    def __iter__(self):
        return (k for k in self.keys())

    def __str__(self):
        obj = _get(self, SLOT)
        return str(obj)

    def __len__(self):
        obj = _get(self, SLOT)
        return len(obj)

    def __call__(self, *args, **kwargs):
        obj = _get(self, SLOT)
        return obj(*args, **kwargs)


register_data(DataObject)


def get_keys(obj):
    """
    RETURN keys OF obj, AS IF DATA
    """
    while isinstance(obj, (DataObject, Data)):
        obj = _get(obj, SLOT)

    if isinstance(obj, dict):
        return obj.keys()

    try:
        return obj.__dict__.keys()
    except Exception:
        pass

    _type = _get(obj, CLASS)
    keys = _known_fields.get(_type)
    if keys is not None:
        return keys

    try:
        keys = _known_fields[_type] = _type.__slots__
        return keys
    except Exception:
        pass

    keys = _known_fields[_type] = tuple(
        k
        for k in dir(_type)
        if k not in ignored_attributes
        and getattr(_type, k).__class__.__name__ in ["member_descriptor", "getset_descriptor"]
    )
    return keys


def object_to_data(v):
    try:
        if is_null(v):
            return Null
    except Exception:
        pass

    if is_primitive(v):
        return v

    _class = _get(v, CLASS)
    if _class in (dict, OrderedDict):
        m = _new(Data)
        _set(m, SLOT, v)
        return m
    elif _class in (tuple, list):
        return list_to_data(v)
    elif _class in (Data, DataObject, FlatList, NullType):
        return v
    elif _class in generator_types:
        return (to_data(vv) for vv in v)
    elif is_known_data_type(_class):
        return DataObject(v)
    else:
        return v


class DataClass(object):
    """
    ALLOW INSTANCES OF class_ TO ACT LIKE dicts
    ALLOW CONSTRUCTOR TO ACCEPT @override
    """

    def __init__(self, _class):
        register_type(_class)
        self.class_ = _class
        self.constructor = _class.__init__

    def __call__(self, *args, **kwargs):
        settings = to_data(kwargs).settings

        params = get_function_arguments(self.constructor)[1:]
        func_defaults = get_function_defaults(self.constructor)
        if not func_defaults:
            defaults = {}
        else:
            defaults = {k: v for k, v in zip(reversed(params), reversed(func_defaults))}

        ordered_params = dict(zip(params, args))

        output = self.class_(**params_pack(params, ordered_params, kwargs, settings, defaults))
        return DataObject(output)


def params_pack(params, *args):
    settings = {}
    for a in args:
        for k, v in a.items():
            k = str(k)
            if k in settings:
                continue
            settings[k] = v

    output = {str(k): from_data(settings[k]) for k in params if k in settings}
    return output


export("mo_dots.lists", object_to_data)
export("mo_dots.datas", object_to_data)
export("mo_dots.datas", DataObject)
export("mo_dots.datas", get_keys)
